#!/usr/local/bin/perl

# randomPairs.pl version 1.1
#
# A script to generate a list of random pairs of word senses extracted
# from WordNet.
#
# Copyright (C) 2004
#
# Jason Michelizzi, University of Minnesota Duluth
# mich0212 at d.umn.edu
#
# Ted Pedersen, University of Minnesota Duluth
# tpederse at d.umn.edu
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


use strict;
use warnings;

use WordNet::QueryData;

use File::Spec;

use Getopt::Long;

our ($opt_wnpath, $opt_num, $opt_help, $opt_version, $opt_outfile);
our ($opt_noun, $opt_verb, $opt_adj, $opt_adv);

my $result = GetOptions ("wnpath=s", "num|n=i", "help", "version", "outfile=s",
			 "noun=f", "verb=f", "adj=f", "adv=f");

unless ($result) {
  showUsage ();
  exit (1);
}

if ($opt_help) {
  showHelp ();
  exit (0);
}

if ($opt_version) {
  showVersion ();
  exit (0);
}

$opt_num = 128 unless $opt_num;

fixRatios();

my $wnpath;

{
  my $rootdir = File::Spec->rootdir;
  my $wnhome;

  if ($^O !~ /Win/i) {
    $wnhome = File::Spec->catdir ($rootdir, 'usr','local', 'WordNet-2.0');
  }
  else {
    $wnhome = File::Spec->catdir ($rootdir, 'WordNet', '2.0');
  }
  my $wnsearchdir = 'dict';

  $wnhome = $ENV{WNHOME} if defined $ENV{WNHOME};
  $wnsearchdir = $ENV{WNSEARCHDIR} if defined $ENV{WNSEARCHDIR};

  $wnpath = File::Spec->catfile ($wnhome, $wnsearchdir);
  $wnpath = $opt_wnpath if $opt_wnpath;
}


print STDERR "Loading WordNet::QueryData... ";
my $wn = new WordNet::QueryData ($wnpath);
$wn or die "failed to load WordNet::QueryData";
print STDERR "done.\n";

my ($fhn, $fhv, $fha, $fhr);
open $fhn, "$wnpath/data.noun" or die "could not open data.noun: $!";
open $fhv, "$wnpath/data.verb" or die "could not open data.verb: $!";
open $fha, "$wnpath/data.adj" or die "could not open data.adj: $!";
open $fhr, "$wnpath/data.adv" or die "could not open data.adv: $!";

my %fhs = (n => $fhn, v => $fhv, a => $fha, r => $fhr);
my %pos = (0 => 'n', 1 => 'v', 2 => 'a', 3 => 'r');

my %lines = getLinesByPos(\%fhs);

my $fh = *STDOUT;
if ($opt_outfile) {
  undef $fh;
  open $fh, ">$opt_outfile" or die "Cannot open $opt_outfile: $!";
}

for my $i (1..$opt_num) {
  my @pair;
  for my $j (0..1) {
    my $t = int rand 100;

    my $posnum;
    if ($t < $opt_noun) {
      $posnum = 0;
    }
    elsif ($t < $opt_verb) {
      $posnum = 1;
    }
    elsif ($t < $opt_adj) {
      $posnum = 2;
    }
    else {
      $posnum = 3;
    }

    my $pos = $pos{$posnum};
    my $fh = $fhs{$pos};

    seek ($fh, 1740, 0) or die "Failed to seek properly: $!";

    my $line_num = int rand ($lines{$pos});

    <$fh> for (1..$line_num);
    my $line = <$fh>;
    my ($offset) = split /\s+/, $line;
    my $sense = $wn->getSense ($offset, $pos);

    my @syns = $wn->querySense ($sense, "syns");
    $sense = $syns[int rand scalar @syns];

    defined $sense or die "Undefined sense";

    $pair[$j] = $sense;
  }
  print $fh "$pair[0] $pair[1]\n";
}

close $fhs{$_} foreach (qw/n v a r/);


###### subs start here:

# gets the number of synsets for each part of speech
sub getLinesByPos {
  # for WordNet 2.0, the line counts should be:
  # 79718 - 29 for nouns
  # 13537 - 29 for verbs
  # 18592 - 29 for adjectives
  # 3693 - 29 for adverbs
  # first synset is at 00001740 for all POS.
  #(n => 79689, v => 12696, a => 18563, r => 3664);

  my $fhs = shift;
  my %rethash;

  while (my ($pos, $fh) = each %$fhs) {
    my $lines = 0;

    seek $fh, 0, 0 or die "Failed to 'seek': $!";

    while (<$fh>) {
      next if /^\s+/;
      $lines++;
    }

    seek ($fh, 0, 0) or die "Failed to 'seek': $!";

    $rethash{$pos} = $lines;
  }
  return %rethash;
}

sub fixRatios
{
  # The following comes from wnstats(7).
  # Unique word-sense pairs in WordNet 2.0:
  # Noun:  141,690  69.7482%
  # Verb:   24,632  12.1253%
  # Adj:    31,015  15.2674%
  # Adv:     5,808   2.8590%
  # ------------------------
  # Total: 203,145  99.9999%

  #my $nounratio = .697482;
  #my $verbratio = .121253;
  #my $adjratio = .152674;
  #my $advratio = .028590;

  my $total_assigned = 0;
  my $i = 0;
  do {$total_assigned += $opt_noun, $i++} if defined $opt_noun;
  do {$total_assigned += $opt_verb, $i++} if defined $opt_verb;
  do {$total_assigned += $opt_adj, $i++} if defined $opt_adj;
  do {$total_assigned += $opt_adv, $i++} if defined $opt_adv;

  if ($total_assigned > 100.0) {
    print "Bad combination of ratios.  The sum of the ratios must be\n";
    print "less-than or equal-to to 100.\n";
    showUsage ();
    exit (1);
  }

  my $j = (100 - $total_assigned);
  $j /= (4 - $i) if $i < 4;
  $opt_noun = $j unless defined $opt_noun;
  $opt_verb = $j unless defined $opt_noun;
  $opt_adj = $j unless defined $opt_adj;
  $opt_adv = $j unless defined $opt_adv;

  $opt_verb += $opt_noun;
  $opt_adj += $opt_verb;
  $opt_adv += $opt_adj;
}


sub showVersion {
  print "randomPairs.pl version 1.1\n";
  print "Copyright (C) 2004, Jason Michelizzi and Ted Pedersen\n\n";
  print "randomPairs.pl comes with ABSOLUTELY NO WARRANTY.  This is free\n";
  print "software, and you are welcome to redistribute it under certain\n";
  print "conditions.  See the file GPL that came with this distribution.\n";
}

sub showUsage {
  print "Usage: randomPairs.pl [--wnpath=PATH]  [--num=NUM] [--outfile=FILE] [--noun=REAL] [--verb=REAL] [--adj=REAL] [--adv=REAL] | --help | --version]\n";
}

sub showHelp {
  showUsage();

  print "--wnpath=PATH   Specify the location of WordNet data files. The\n";
  print "                 default on Unix is /usr/local/WordNet-2.0/dict\n";
  print "--num=NUM       Specify the number of word pairs to generate\n";
  print "--outfile=FILE  Specify the file to which output should be sent.\n";
  print "                 By default, output is sent to standard output.\n";
  print "--noun=REAL\n";
  print "--verb=REAL\n";
  print "--adj=REAL\n";
  print "--adv=REAL      Specify the proportion of words that should be nouns\n";
  print "                 verbs, adjectives, and adverbs respectively.  The\n";
  print "                 values for these options must be real numbers between\n";
  print "                 0 and 100, and the sum of the values should be 100\n";
  print "--help          Show this help message\n";
  print "--version       Show version information\n";
}

__END__

=head1 NAME

randomPairs.pl -- generate a list of pairs of random word senses

=head1 SYNOPSIS

randomPairs.pl [--wnpath=PATH] [--num=NUM] [--outfile=FILE] | --help | --version

=head1 DESCRIPTION

=over

=item --wnpath=PATH

Specifies the path to the WordNet data files.  The default for Unix is
/usr/local/WordNet-2.0/dict and the default for Windows is C:\WordNet\2.0\dict.
You can also use the WNHOME and WNSEARCHDIR environment variables to specify
the location of WordNet data files.  The program assumes the data files are in
WNHOME/WNSEARCHDIR.  WNHOME is assumed to be "/usr/local/WordNet-2.0" and 
WNSEARCHDIR is assumed to be "dict".

=item --num=NUM

Specifies the number of word pairs to generate.  The default is 128.

=item --outfile=FILE

Specifies the file to which output should be sent.  If the option is not
specified then output is sent to the standard output.

=item --help

Shows a detailed help message.

=item --version

Shows version information

=item --noun=REAL --verb=REAL --adj=REAL --adv=REAL

Specifies the proportion of words that should be nouns, verbs, adjectives, and
adverbs respecively.  The value for each option must be a real number between
0 and 100 (inclusive).  The sum for all the values should be 100 if all four
options are specified.  One or more of the options is not specified, then the
sum should be less-than or equal-to 100.  If one or more options is omitted,
then the omitted options are assigned a value such that the value for the
omitted option i is

   value_i = (100 - Sum) / number_omitted

where Sum is the sum of the values for all the options that are specified.

=back

=head1 BUGS

None

=head1 AUTHORS

Jason Michelizzi <mich0212 at d.umn.edu> and Ted Pedersen <tpederse at d.umn.edu>.

=head1 COPYRIGHT AND LICENSES

randomPairs.pl version 1.1

Copyright (C) 2004, Jason Michelizzi and Ted Pedersen

randomPairs.pl comes with ABSOLUTELY NO WARRANTY.  This is free software,
and you are welcome to redistribute it under certain conditions.  See the
file GPL that came with this distribution.

=cut
